#pylint: disable-all
""" Python bindings for running programs from the SimpactCyan package.

This package supplies a class `PySimpactCyan`, which can be used to run
simulations using the SimpactCyan programs. When used as a standalone
python program, it can be used to list a full config file based on just$
a few fields which override the defaults.
"""

from __future__ import print_function
import sys
import pprint
import copy
import subprocess
import json
import os
import tempfile
import shutil
import random
import time
import platform

# This helper function executes 'executable', for example
#
#     [ 'simpact-cyan-opt', '--showconfigoptions' ]
#
# to retrieve all config options in JSON format. It then expands the
# options which are of type 'distTypes' into the available 1D distribution
# types listed in the JSON config

def _getExpandedSettingsOptions(executable):

    with open(os.devnull, "w") as nullFile:
        proc = subprocess.Popen(executable, stdout=subprocess.PIPE, stderr=nullFile)
        jsonData, unusedErr = proc.communicate()
        proc.wait()

    jsonData = jsonData.decode("utf-8") # Needed for python3

    configOptions = json.loads(jsonData)
    configNames = configOptions["configNames"]
    
    distTypesNames = [ ]
    for n in configOptions:
        if n != "configNames":
            distTypesNames.append(n)

    possibleDistNames = { }
    for n in distTypesNames:
        possibleDistNames[n] = [ t for t in configOptions[n] ]

    # Change the config entries which have a 'distTypes' etc setting

    newConfig = copy.deepcopy(configNames)
    for n in configNames:

        params = configNames[n]['params']
        for i in range(len(params)):
            p = params[i] 
            pName = p[0]
            pValue = p[1]

            if pValue in distTypesNames:

                distTypes = configOptions[pValue]

                defaultDistName = "fixed"
                defaultDistParams = None
                if len(p) == 3: # The third parameter are the defaults
                    defaultDistOptions = p[2]
                    defaultDistName = defaultDistOptions[0]
                    defaultDistParams = defaultDistOptions[1]

                # Adjust the entry in 'newConfig' to reflect the default distribution name
                newConfig[n]['params'][i] = [ pName + ".type", defaultDistName, possibleDistNames[pValue] ]

                # Add entries to newConfig for all possible distributions
                for distName in distTypes:
                    distParams = distTypes[distName]['params']

                    newConfName = n + "_" + str(i) + "_" + distName
                    newConfig[newConfName] = { 'depends': [ n, pName + ".type", distName ] }

                    if defaultDistName == distName and defaultDistParams:
                        # Default parameters are present for this distribution,
                        # merge them

                        defaultParamMap = { }
                        for dp,dv in defaultDistParams:
                            defaultParamMap[dp] = dv

                        modParams = copy.deepcopy(distParams)
                        for paramPair in modParams:
                            dp = paramPair[0]
                            if dp in defaultParamMap:
                                paramPair[1] = defaultParamMap[dp]

                        newConfig[newConfName]['params'] = [ [ pName + "." + distName + "." + p[0] ] + p[1:] for p in modParams ]
                    else:
                        # No specific defaults, just use the info from the "distTypes" data structure
                        newConfig[newConfName]['params'] = [ [ pName + "." + distName + "." + p[0] ] + p[1:] for p in distParams ]

                    newConfig[newConfName]['info'] = distTypes[distName]['info']

    #pprint.pprint(newConfig)
    #sys.exit(-1)
    return newConfig

# userConfig is a map of key/value pairs, for each configuration option as
# will be present in the config file
#
# cfg is actually an entry of configNames, and configNames is present as
# well in case this entry has a dependency, which then needs to be checked
# first
#
# requiredKeys will be a map of keys than need to be set in the final config
# file, either mapped to None just to indicate that it needs to be present,
# or mapped to a list of possible values
def _processConfigPart(cfg, userConfig, configNames, requiredKeys):

    params = cfg['params']
    deps = cfg['depends']

    # Check if we need to process dependencies first
    if deps is not None:

        depObjName = deps[0]
        depObj = configNames[depObjName]
        depKey = deps[1]
        depVal = deps[2]
        
        #print "processConfigPart", depObjName
        #pprint.pprint(depObj)
        if not _processConfigPart(depObj, userConfig, configNames, requiredKeys):
            # Parent dependency not fulfilled, so this one isn't either
            return False
        #print "done: processConfigPart", depObjName

        if not depKey in userConfig:
            pprint.pprint(userConfig)
            raise Exception("Key %s was not set" % depKey)
        
        if userConfig[depKey] != depVal:
            return False # Dependency not fulfilled

        for k in params:
            if len(k) == 3:
                requiredKeys[k[0]] = k[2]
            else:
                requiredKeys[k[0]] = None

    for p in params:

        key = p[0]
        val = p[1]

        if len(p) == 3:

            requiredKeys[key] = p[2]
        else:
            requiredKeys[key] = None

        # See if we should check defaults
        if not key in userConfig: 
            #if val is None:
            #    raise Exception("Key %s is not set" % key)

            userConfig[key] = val
        
    return True

def createConfigLines(executable, inputConfig, checkNone = True, ignoreKeys = [ ]):

    def getValidNames(strList):
        r = "['" + strList[0] + "'"
        for i in range(1,len(strList)):
            r += ", '" + strList[i] + "'"
        r += "]"
        return r

    userConfig = copy.deepcopy(inputConfig)
    configNames = _getExpandedSettingsOptions(executable)

    requiredKeys = { }

    for n in configNames:
        cfg = configNames[n]
        _processConfigPart(cfg, userConfig, configNames, requiredKeys)

    for k in userConfig:
        if not k in requiredKeys:
            raise Exception("Encountered unknown key %s" % k)

        val = userConfig[k]

        possibleValues = requiredKeys[k]
        if possibleValues is not None:
            if not val in possibleValues:
                raise Exception("Value '%s' for key %s is not allowed, should be one of %s" % (val, k, possibleValues))

        if checkNone:
            if val is None:
                raise Exception("Key %s is not set" % k)

    # Display the final config file

    lines = [ ]
    unusedlines = [ ]
    # In principle this should contain the same info as userConfig at the end,
    # but we'll introduce some ordering here so we can feed it back to R in a better
    # way
    resultingConfig = [ ] 

    names = [ key for key in configNames ]
    names.sort()
    for key in names:
        deps = configNames[key]["depends"]
        params = configNames[key]["params"]
        info = configNames[key]["info"]

        if info:
            info = "\n".join(info)

        usedparams = [ ]
        unusedparams = [ ]
        for p in params:
            k = p[0]
            if k in requiredKeys:
                
                v = userConfig[k]
                ns = 60-len(k)
                k += " "*ns

                if len(p) == 3: # Limited number of possibilities
                    usedparams.append("# Valid values are: " + getValidNames(p[2]))

                if v is None:
                    usedparams.append("%s = " % k)
                elif type(v) == float:
                    usedparams.append("%s = %.15g" % (k, v))
                elif type(v) == int:
                    usedparams.append("%s = %d" % (k, v))
                else:
                    usedparams.append("%s = %s" % (k, str(v)))

                idx = len(resultingConfig)+1

                if v is None:
                    resultingConfig.append((idx, p[0], ""))
                else:
                    resultingConfig.append((idx, p[0], v))

            else:
                unusedparams.append("# " + p[0])

        if usedparams:
            if deps:
                lines += [ "# The following depends on %s = %s" % (deps[1], deps[2]) ]

            if info:
                lines += [ "# " + l for l in info.splitlines() ]

            lines += usedparams
            lines += [ "" ]

        if unusedparams:
            if deps:
                unusedlines += [ "# The following depends on %s = %s" % (deps[1], deps[2]) ]

            unusedlines += unusedparams
            unusedlines += [ "#" ]

    introlines = [  "# The configuration file format is quite straightforward, it is just a set of",
                    "# 'key = value' lines. Lines that start with '#' are treated as comments and",
                    "# are ignored.",
                    "#",
                    "# If the key starts with a dollar ('$') sign, the rest of the key is ",
                    "# considered to be the name of a variable, which may be used later on in the",
                    "# file. To use such a variable in a specified value, you need to surround",
                    "# the variable name with '${' and '}'. For example, one could write:",
                    "#",
                    "#     $PREFIX = MyCustomPrefix",
                    "#     logsystem.outfile.logevents = ${PREFIX}-output.log",
                    "#",
                    "# and the file used will have the name 'MyCustomPrefix-output.log'.",
                    "#",
                    "# In the same way, environment variables can be used, and, in fact, ",
                    "# environment variables will take precedence over these internal variables.",
                    "# This way, it is easy to change the content of these variables on the command",
                    "# line",
                    "#",
                    "# Note that no calculations can be performed in this file anymore, so instead",
                    "# of writing 1.0/2.0, you'd need to write 0.5 for example.",
                    "" ]

    return (userConfig, introlines + unusedlines + [ "" ] + lines, resultingConfig)

def _replaceVariables(value, variables):

    newValue = ""

    done = False
    prevIdx = 0
    while not done:
        idx = value.find("${", prevIdx)
        if idx < 0:
            done = True
        else:
            nextIdx = value.find("}", idx)
            if nextIdx < 0:
                done = True
            else:
                key = value[idx+2:nextIdx]
                if key in variables:
                    newValue += variables[key]
                    prevIdx = nextIdx+1
                else:
                    newValue += value[prevIdx:nextIdx+1]
                    prevIdx = nextIdx + 1
    
    newValue += value[prevIdx:]

    return newValue.strip()

def _getSimpactPathBasedOnModule():
    
    def endSlash(dirName):
        if not dirName.endswith(os.sep): # Make sure it ends with "/" or "\"
            dirName += os.sep
        return dirName

    possiblePaths = [ ]

    moduleName = __name__
    paths = sys.path
    for p in paths:
        if os.path.isdir(p):
            full = os.path.join(p, moduleName + ".py")
            if os.path.exists(full): # Ok, our simpact module exists in this directory
                full = os.path.abspath(full)
                dirName = os.path.dirname(full) 
                baseName = os.path.basename(dirName) # should be 'python'
                if baseName.lower() == "python":
                    dirName = endSlash(os.path.dirname(dirName))
                    possiblePaths.append(dirName)

                if baseName.lower() == "site-packages":
                    dirName = os.path.dirname(os.path.dirname(dirName))
                    possiblePaths.append(endSlash(os.path.join(dirName, "Library", "share", "simpact-cyan")))
                    dirName = os.path.dirname(dirName)
                    possiblePaths.append(endSlash(os.path.join(dirName, "share", "simpact-cyan")))

    #print("Possible paths:")
    #print(possiblePaths)

    return possiblePaths

class PySimpactCyan(object):
    """ This class is used to run SimpactCyan based simulations."""

    def __init__(self):
        self._execPrefix = "simpact-cyan"
        self._dataDirectory = self._findSimpactDataDirectory()
        self._execDir = self._findSimpactDirectory()

    def setSimpactDirectory(self, dirName):
        """ Sets the directory in which the simpact binaries were installed to `dirName`"""
        self._execDir = os.path.abspath(dirName)

    def setSimpactDataDirectory(self, dirName):
        self._dataDirectory = dirName

    def _findSimpactDataDirectory(self):
        paths = [ ]

        if "SIMPACT_DATA_DIR" in os.environ:
            paths += [ os.environ["SIMPACT_DATA_DIR"] ]

        if platform.system() == "Windows":
            paths += [ "C:\\Program Files (x86)\\SimpactCyan\\data\\", "C:\\Program Files\\SimpactCyan\\data\\" ]
        else:
            paths += [ "/usr/share/simpact-cyan/", "/usr/local/share/simpact-cyan/" ]
            paths += [ "/Applications/SimpactCyan.app/Contents/data/" ]

        for p in _getSimpactPathBasedOnModule():
            paths.append(p)
            p2 = os.path.join(p, "data")
            if not p2.endswith(os.sep):
                p2 += os.sep

            paths.append(p2)

        for p in paths:
            f = os.path.join(p, "sa_2003.csv")
            if os.path.exists(f):
                print("Setting data directory to", p)
                return p

        print("Warning: can't seem to find the simpact data directory")
        return None

    def _findSimpactDirectory(self):
        with open(os.devnull, "w") as DEVNULL:
            paths = [ ]
            if platform.system() == "Windows":
                paths += [ "C:\\Program Files (x86)\\SimpactCyan", "C:\\Program Files\\SimpactCyan" ]
            else:
                paths += [ "/Applications/SimpactCyan.app/Contents/bin" ]

            for p in _getSimpactPathBasedOnModule():
                paths.append(p)
                p2 = os.path.join(p, "bin")
                if not p2.endswith(os.sep):
                    p2 += os.sep

                paths.append(p2)

            exe = "simpact-cyan-release" # This should always exist

            # First see if we can run the executable without a full path
            try:
                subprocess.call([ exe ], stderr=DEVNULL, stdout=DEVNULL)
                return None
            except:
                pass

            # Then try some predefined paths
            for p in paths:
                try:
                    subprocess.call([ os.path.join(p,exe) ], stderr=DEVNULL, stdout=DEVNULL)
                    print("Simpact executables found in %s" % p)
                    return p
                except:
                    pass

        print("Warning: can't seem to find a way to run the simpact executables")
        return None

    def setSimulationPrefix(self, prefix):

        if not prefix:
            raise Exception("No valid simulation prefix specified")

        with open(os.devnull, "w") as DEVNULL:
            try:
                p = self._getExecPath(testPrefix = prefix)
                subprocess.call( [ p ], stderr=DEVNULL, stdout=DEVNULL)
                self._execPrefix = prefix
            except Exception as e:
                raise Exception("Unable to use specified prefix '%s' (can't run '%s')" % (prefix, p))

    def _getExecPath(self, opt = True, release = True, testPrefix = None):

        fullPath = testPrefix if testPrefix else self._execPrefix
        fullPath += "-"
        fullPath += "release" if release else "debug"

        if self._execDir is not None:
            fullPath = os.path.join(self._execDir, fullPath)
        return fullPath

    def runDirect(self, configFile, parallel = False, opt = True, release = True, outputFile = None, seed = -1, destDir = None, quiet = False):

        fullPath = self._getExecPath(opt, release)
        parallelStr = "1" if parallel else "0"

        if type(opt) == bool:
            algoStr = "opt" if opt else "simple"
        else:
            algoStr = str(opt)

        if destDir is None:
            destDir = os.path.abspath(os.path.dirname(configFile))
    
        closeOutput = False
        origDir = os.getcwd()
        try:
            os.chdir(destDir)

            if outputFile is not None:
                if os.path.exists(outputFile):
                    raise Exception("Want to write to output file '%s', but this already exists" % outputFile)

                f = open(outputFile, "w+t")
                closeOutput = True
            else:
                f = tempfile.TemporaryFile(mode='w+t')

            newEnv = copy.deepcopy(os.environ)
            if seed >= 0:
                newEnv["MNRM_DEBUG_SEED"] = str(seed)

            if self._dataDirectory is not None:
                newEnv["SIMPACT_DATA_DIR"] = str(self._dataDirectory)
            
            if not quiet:
                print("Results will be stored in directory '%s'" % os.getcwd())
                print("Running simpact executable '{}' ...".format(fullPath))

            proc = subprocess.Popen([fullPath, configFile, parallelStr, algoStr], stdout=f, stderr=f, cwd=os.getcwd(), env=newEnv)
            try:
                proc.wait() # Wait for the process to finish
            except:
                try: 
                    proc.kill()
                except:
                    pass
                raise

            f.flush()
            f.seek(0)
            lines = f.readlines()

            if not quiet:
                print("Done.")
                print()

                # Show contents of output file or temporary file on screen

                for l in lines:
                    sys.stdout.write(l)
                sys.stdout.flush()

            if proc.returncode != 0:
                raise Exception(self._getProgramExitError(lines, proc.returncode))

        finally:
            os.chdir(origDir)

    def _getProgramExitError(self, lines, code):
        lines = [ l.strip() for l in lines if l.strip() ]

        # Look for 'FATAL ERROR'
        for i in range(len(lines)):
            l = lines[i]
            if l == "FATAL ERROR:" and i+1 < len(lines):
                return lines[i+1]

        # Look for last line before 'UNEXPECTED TERMINATION OF PROGRAM!'
        for i in range(len(lines)):
            l = lines[i]
            if l == "UNEXPECTED TERMINATION OF PROGRAM!" and i > 0:
                return lines[i-1]

        return "Program exited with an error code ({})".format(code)

    def _createConfigLines(self, inputConfig, checkNone = True, ignoreKeys = []):
        executable = [ self._getExecPath(), "--showconfigoptions" ]
        return createConfigLines(executable, inputConfig, checkNone, ignoreKeys)

    def _checkKnownKeys(self, keyList):
        executable = [ self._getExecPath(), "--showconfigoptions" ]

        configNames = _getExpandedSettingsOptions(executable)

        allKnownKeys = [ ]
        for n in configNames:
            paramList = configNames[n]["params"]
            paramKeys = [ ]
            for p in paramList:
                paramKeys.append(p[0])

            allKnownKeys += paramKeys

        for k in keyList:
            if not k in allKnownKeys:
                raise Exception("Encountered unknown key '%s'" % k)

    def getConfiguration(self, config, show = False):
        # Make sure config is a dict
        if not config:
            config = { }

        ignoreKeys = [ "population.agedistfile" ]

        finalConfig, lines, sortedConfig = self._createConfigLines(config, False, ignoreKeys)
        lines.append('')

        if show:
            sys.stdout.write('\n'.join(lines))

        return sortedConfig

    def showConfiguration(self, config):
        self.getConfiguration(config, True)

    def _getID(self, identifierFormat):

        def getRandomChar():
            chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
            pos = int(random.random()*len(chars)) % len(chars)
            return chars[pos]

        t = time.gmtime()
        pid = os.getpid()
        simType = self._execPrefix

        identifier = ""
        prevPos = 0
        pos = identifierFormat.find("%")
        while pos >= 0:
            if pos < len(identifierFormat)-1 and identifierFormat[pos+1] in [ '%', 'T', 'y', 'm', 'd', 'H', 'M', 'S', 'p', 'r' ]:

                identifier += identifierFormat[prevPos:pos]
                prevPos = pos+2

                n = identifierFormat[pos+1]
                if n == '%':
                    identifier += "%"
                elif n == 'T':
                    identifier += simType
                elif n == 'y':
                    identifier += "%d" % t.tm_year
                elif n == 'm':
                    identifier += "%02d" % t.tm_mon
                elif n == 'd':
                    identifier += "%02d" % t.tm_mday
                elif n == 'H':
                    identifier += "%02d" % t.tm_hour
                elif n == 'M':
                    identifier += "%02d" % t.tm_min
                elif n == 'S':
                    identifier += "%02d" % t.tm_sec
                elif n == 'p':
                    identifier += "%d" % pid
                elif n == 'r':
                    identifier += getRandomChar()
                else:
                    raise Exception("Internal error: unexpected identifier format '%s'" % n)

                pos = identifierFormat.find("%", pos+2)
            else:
                # No need to adjust prevPos
                pos = identifierFormat.find("%", pos+1)

        identifier += identifierFormat[prevPos:]

        return identifier

    def _getFromConfigOrSetDefault(self, config, key, defaultValue):

        try:
            v = config[key]
        except:
            v = defaultValue
            config[key] = defaultValue

        return v

    def _writeDataFrame(self, f, data, dataName):
        nrows, ncols = data.shape

        colNames = data.columns
        if len(colNames) == ncols:
            f.write(",".join(['"{}"'.format(colNames[c]) for c in range(ncols)]))
            f.write("\n")

        matrix = data.as_matrix()
        self._writeMatrix(f, matrix, dataName, False)

    def _writeMatrix(self, f, data, dataName, writeColumnNames = True):
        nrows = len(data)
        ncols = -1
        for r in range(nrows):

            if ncols < 0:
                ncols = len(data[r])

            if len(data[r]) != ncols:
                raise Exception("Error processing row %d for data '%s': expecting %d columns, but got %d" % (r+1, dataName, ncols, len(data[r])))

            if writeColumnNames and r == 0:
                colNames = [ '"Col{}"'.format(c+1) for c in range(ncols) ]
                colNamesStr = ",".join(colNames)
                f.write(colNamesStr + "\n")

            entries = [ "%.15g" % data[r][c] for c in range(ncols) ]
            entriesStr = ",".join(entries)
            f.write(entriesStr)
            f.write("\n")

    def _writeDataFile(self, destDir, fileName, dataName, data):
        fullPath = os.path.join(destDir, fileName)
        if os.path.exists(fullPath):
            raise Exception("Error while writing data file for '%s': file '%s' already exists" % (dataName, fileName))

        with open(fullPath, "wt") as f:

            isDataFrame = False
            try:
                s = data.columns # Assume that it's a pandas dataframe if this exists
                isDataFrame = True
            except:
                pass

            if isDataFrame:
                self._writeDataFrame(f, data, dataName)
            else: # Not a dataframe
                self._writeMatrix(f, data, dataName)

    def _toFileName(self, s):
        ret = ""
        for c in s:
            if c.isalnum():
                ret += c
            else:
                ret += "_"
        return ret

    def run(self, config, destDir, agedist = None, parallel = False, opt = True, release = True, seed = -1, 
            interventionConfig = None, dryRun = False, identifierFormat = "%T-%y-%m-%d-%H-%M-%S_%p_%r%r%r%r%r%r%r%r-",
            dataFiles = { }, quiet = False):

        if not destDir:
            raise Exception("A destination directory must be specified")

        # Make sure config is a dict
        if not config:
            config = { }

        originalConfig = copy.deepcopy(config)

        idStr = self._getID(identifierFormat)
        if not idStr:
            raise Exception("The specified identifier format leads to an empty identifier")

        distFile = None
        if agedist is not None:
            if isinstance(agedist, dict):
                distAges = agedist["Age"]

                if "Percent.Male" in agedist:
                    distMalePct = agedist["Percent.Male"]
                elif "Percent Male" in agedist:
                    distMalePct = agedist["Percent Male"]
                else:
                    raise Exception("Error in age distribution: Key for male percentage must be 'Percent.Male' or 'Percent Male'")

                if "Percent.Female" in agedist:
                    distFemalePct = agedist["Percent.Female"]
                elif "Percent Female" in agedist:
                    distFemalePct = agedist["Percent Female"]
                else:
                    raise Exception("Error in age distribution: Key for female percentage must be 'Percent.Female' or 'Percent Female'")

                if len(distAges) != len(distMalePct) or len(distAges) != len(distFemalePct):
                    raise Exception("Not all columns of the 'agedist' variable seem to have the same length")

                distFile = "%sagedist.csv" % idStr
                config["population.agedistfile"] = distFile

            else: # Assume we're referring to a file

                config["population.agedistfile"] = str(agedist)


        # Intervention event stuff

        intTimes = None
        intBaseFile = None
        ivIDs = []
        if interventionConfig:
            # Lets make sure we order the intervention times

            intTimes = [ ]
            if type(interventionConfig) == dict:

                for iv in interventionConfig:
                    t = float(interventionConfig[iv]["time"])
                    del interventionConfig[iv]["time"]
                    intTimes.append( (t, interventionConfig[iv]) )

            else: # assume it's a list
                for iv in interventionConfig:
                    t = float(iv["time"])
                    del iv["time"]
                    intTimes.append( (t, iv) )

            intTimes.sort() # Make sure it's sorted on time, interpreted as a real number

        if intTimes:

            config["intervention.enabled"] = "yes"            

            isFirstTime = True
            ivTimeString = ""
            ivIDString = ""
            count = 1
            for (t,iv) in intTimes:
                if not isFirstTime:
                    ivTimeString += ","
                    ivIDString += ","
                
                ivTimeString += str(t)
                ivIDString += str(count)
                ivIDs.append(str(count))

                isFirstTime = False
                count += 1
                
            config["intervention.times"] = ivTimeString
            config["intervention.fileids"] = ivIDString

            intBaseFile = "%sinterventionconfig_%%.txt" % idStr
            config["intervention.baseconfigname"] = intBaseFile

        if os.path.exists(destDir):
            # Check that we're actually dealing with a directory
            if not os.path.isdir(destDir):
                raise Exception("Specified destination directory '%s' exists but does not seem to be a directory" % destDir)

        else:
            # Create the directory
            if not quiet:
                print("Specified destination directory '%s' does not exist, creating it" % destDir)
            os.makedirs(destDir)

        # Replace things that start with "data:" by "${SIMPACT_INDATA_PREFIX}"
        dataPrefix = "data:"
        for c in config:
            v = config[c]
            if str(v).startswith(dataPrefix):
                v = "${SIMPACT_INDATA_PREFIX}" + self._toFileName(v[len(dataPrefix):]) + ".csv"
                config[c] = v

        # Here, the actual configuration file lines are created
        finalConfig, lines, notNeeded = self._createConfigLines(config, True)

        # Check some paths
        configFile = os.path.abspath(os.path.join(destDir, "%sconfig.txt" % idStr))
        if os.path.exists(configFile):
            raise Exception("Want to write to configuration file '%s', but this already exists" % configFile)

        outputFile = os.path.abspath(os.path.join(destDir, "%soutput.txt" % idStr))
        if os.path.exists(outputFile):
            raise Exception("Want to write to output file '%s', but this already exists" % outputFile)

        if distFile:
            fullDistFile = os.path.abspath(os.path.join(destDir, distFile))
            if os.path.exists(fullDistFile):
                raise Exception("Want to write to age distribution file '%s', but this already exists" % fullDistFile)

        # Write the config file
        with open(configFile, "wt") as f:
            f.write("# Some variables. Note that if set, environment variables will have\n")
            f.write("# Precedence.\n")
            f.write("$SIMPACT_OUTPUT_PREFIX = %s\n" % idStr)
            if dataFiles:
                f.write("$SIMPACT_INDATA_PREFIX = %s\n" % (idStr + "data-"))
            if self._dataDirectory:
                f.write("$SIMPACT_DATA_DIR = %s\n" % self._dataDirectory)	
            f.write("\n")

            for l in lines:
                f.write(l + "\n")
            f.close()

        if distFile:
            # Write the age distribution file
            with open(fullDistFile, "wt") as f:
                f.write("Age,Percent Male,Percent Female\n")
                for i in range(len(distAges)):
                    f.write("%g,%g,%g\n" % (distAges[i], distMalePct[i], distFemalePct[i]))
                f.close()

        # write intervention config files
        if intTimes:
            for tIdx in range(len(intTimes)):

                t = intTimes[tIdx][0]
                iv = intTimes[tIdx][1]

                # With the current approach, the best we can do is to check for keys that
                # are never used in a config file
                self._checkKnownKeys([ name for name in iv ])

                fileName = intBaseFile.replace("%", ivIDs[tIdx])
                fileName = os.path.join(destDir, fileName)
                with open(fileName, "w") as f:
                    for k in iv:
                        f.write("%s = %s\n" % (k,iv[k]))
                    f.close()

        # write data files

        #pprint.pprint(dataFiles)
        if dataFiles:
            for d in dataFiles:
                self._writeDataFile(destDir, idStr + "data-" + self._toFileName(d) + ".csv", d, dataFiles[d])

        # Set environment variables (if necessary) and start executable

        if not dryRun:
            if not quiet:
                print("Using identifier '%s'" % idStr)
            self.runDirect(configFile, parallel, opt, release, outputFile, seed, destDir, quiet)

        # Create the return structure
        results = { }
        replaceVars = { }
        
        # These are things that should be replaced
        if self._dataDirectory: 
            replaceVars["SIMPACT_DATA_DIR"] = self._dataDirectory
        replaceVars["SIMPACT_OUTPUT_PREFIX"] = idStr
        replaceVars["SIMPACT_INDATA_PREFIX"] = os.path.join(destDir, idStr + "data-")
        
        # These are the output log files in a generic way
        outFileSpec = ".outfile."
        for n in finalConfig:

            if outFileSpec in n:
                value = _replaceVariables(finalConfig[n], replaceVars)

                if value:
                    pos = n.find(outFileSpec) + len(outFileSpec)
                    logName = n[pos:]

                    fileName = os.path.join(destDir, value)
                    if os.path.exists(fileName):
                        results[logName] = fileName

        # Also show the 'data:' entries in the returned dictionary
        for n in originalConfig:
            if str(originalConfig[n]).startswith(dataPrefix):
                results[originalConfig[n]] = _replaceVariables(finalConfig[n], replaceVars)

        results["configfile"] = os.path.join(destDir, configFile)
        results["outputfile"] = os.path.join(destDir, outputFile)
        results["id"] = idStr

        if distFile:
            results["agedistfile"] = os.path.join(destDir, distFile)
        else:
            results["agedistfile"] = _replaceVariables(finalConfig["population.agedistfile"], replaceVars)

        # Get simulation time and number of events from output file
        try:
            with open(results["outputfile"], "rt") as f:
                lines = f.readlines()

            simTimePrefix = "# Current simulation time is "
            numEvtsPrefix = "# Number of events executed is "

            for l in lines:
                if l.startswith(simTimePrefix):
                    results["simulationtime"] = float(l[len(simTimePrefix):])
                if l.startswith(numEvtsPrefix):
                    results["eventsexecuted"] = int(l[len(numEvtsPrefix):])
        except Exception as e:
            print("WARNING: can't get simulation time or number of events from output file {}: {}".format(results["outputfile"], e))

        return results

def main():

    try:
        executable = [ sys.argv[1], "--showconfigoptions"]
        if len(sys.argv[1:]) != 1:
            raise Exception("Invalid number of arguments")
    except Exception as e:
        print("Error: %s" % str(e), file=sys.stderr)
        print(file=sys.stderr)
        print("Usage: pysimpactcyan.py simpactexecutable", file=sys.stderr)
        sys.exit(-1)

    # Read the input

    userConfig = { }
    line = sys.stdin.readline()
    while line:

        line = line.strip()
        if line:
            parts = [ p.strip() for p in line.split('=') ]
            
            key, value = parts[0], parts[1]
            userConfig[key] = value

        line = sys.stdin.readline()

    # In principle, the 'resultingConfigNotNeeded' should contain the same things
    # as finalConfig, but some ordering was introduced
    (finalConfig, lines, resultingConfigNotNeeded) = createConfigLines(executable, userConfig, False)

    lines.append('')
    sys.stdout.write('\n'.join(lines))

if __name__ == "__main__":
    main()





"""Simpact HIV initialization."""

from whynot.simulators.simpact_hiv import configs
from whynot.simulators.simpact_hiv.simulator import simulate
from whynot.simulators.simpact_hiv.experiments import *



"""Experiments for civil violence simulator."""
import copy
import dataclasses

import numpy as np
import pandas as pd
from whynot.framework import GenericExperiment, parameter
from whynot.simulators import civil_violence
from whynot import utils


def get_experiments():
    """Return all experiments for civil violence."""
    return [RCT]


def run_simulator(agents, config, treatment_assignment, seed):
    """Run the simulator with the specified agents receiving treatment."""
    population = []
    for agent, treated in zip(agents, treatment_assignment):
        agent = copy.deepcopy(agent)
        if treated:
            agent.risk_aversion = .9
        population.append(agent)
    return civil_violence.simulate(population, config, seed)


def run_abm_experiment(agents, treatment_assignments, config, rng, parallelize, show_progress=True):
    """Run an agent-based modeling experiment.

    We run (in parallel) the agent based model for the given treatment
    assignments to simulate a causal experiment. To obtain ground truth
    counterfactuals, we run repeatedly run the simulator with the same random
    seed where (1) no agent is treated [baseline] and (2) where only agent i is
    treated for i=1, 2, ... Thus, the reported ground truth effects are really
    the marginal causal effect.
    """
    baseline_assignments = np.zeros_like(treatment_assignments)
    counterfactual_assignments = []
    for idx in range(len(agents)):
        one_hot = np.zeros_like(treatment_assignments)
        one_hot[idx] = 1.0
        counterfactual_assignments.append(one_hot)

    assignments = [treatment_assignments, baseline_assignments] + counterfactual_assignments

    # Use the same seed for all runs to ensure difference in counterfactual
    # outcome only due to treatment assignment.
    seed = rng.randint(0, 99999)
    parallel_args = [(agents, config, assign, seed) for assign in assignments]

    if parallelize:
        runs = utils.parallelize(run_simulator, parallel_args, show_progress=show_progress)
    else:
        runs = [run_simulator(*args) for args in parallel_args]

    # RCT Outcomes
    observed_outcomes = runs[0].days_active.values

    # Baseline outcomes
    baseline = runs[1].days_active.values

    true_effects = []
    for idx, run in enumerate(runs[2:]):
        effect = run.days_active.values - baseline
        true_effects.append(effect[idx])

    true_effects = np.array(true_effects)

    return observed_outcomes, true_effects


def sample_agents(rng, num_samples, citizen_vision):
    """Generate covariates for each agent in the model."""
    agents = []
    for _ in range(num_samples):
        agent = civil_violence.Agent()
        agent.vision = citizen_vision
        agent.hardship = rng.uniform()
        agent.risk_aversion = .1
        agent.active_threshold = .1
        agent.legitimacy = rng.uniform()
        agents.append(agent)
    return agents


@parameter(name="citizen_vision", default=5, values=list(range(1, 7)),
           description="How many other citizens each agent can see.")
@parameter(name="agent_density", default=0.8, values=np.arange(0.05, 0.95, 0.1),
           description="How densely packed are agents on the grid.")
@parameter(name='cop_fraction', default=0.05, values=[.1, .2, .3],
           description="number of cops = floor(num_samples * cop_fraction)")
@parameter(name="arrest_prob_constant", default=0.9, values=[0.6, 1.2, 2.3, 4.6],
           description="How strong the interaction between agents is.")
@parameter(name="prison_interaction", default=0.4, values=[.01, .05, .1, .15, .2],
           description="Degree to which regime legitimacy is drawn toward the min while in prison")
def run_civil_violence(citizen_vision, agent_density, cop_fraction, arrest_prob_constant,
                       prison_interaction, num_samples=100, seed=None, show_progress=False,
                       parallelize=True):
    """Run an RCT experiment on ABM civil violence model.

    Treatment corresponds to increasing the 'regime_legitimacy'
    belief of an agent. Treatment is assigned randomly, but the
    units interact, and the parameterization of the experiment is
    based upon the degree of interaction.

    Returns:
        - covariates:               np.array, [num_samples, state_dim]
        - treatment:                Binary np.array [num_samples,]
        - outcomes:                 np.array [num_samples,]
        - ground_truth_effect:      sample average treatment effect

    """
    rng = np.random.RandomState(seed)

    # Determine grid size
    num_cops = int(np.floor(cop_fraction * num_samples))
    num_cells = (num_samples + num_cops) / agent_density
    side_length = int(np.ceil(np.sqrt(num_cells)))

    simulator_config = civil_violence.Config(
        grid_height=side_length,
        grid_width=side_length,
        cop_fraction=cop_fraction,
        arrest_prob_constant=arrest_prob_constant,
        prison_interaction=prison_interaction)

    # Generate population of agents
    agents = sample_agents(rng, num_samples, citizen_vision)

    # Assign treatment
    rct_assignments = rng.uniform(size=(num_samples,)) < 0.5

    outcomes, true_effects = run_abm_experiment(
        agents, rct_assignments, simulator_config, rng, parallelize=parallelize,
        show_progress=show_progress)

    covariates = pd.DataFrame(dataclasses.asdict(a) for a in agents).values
    treatment = rct_assignments.astype(np.int64)

    return (covariates, treatment, outcomes), true_effects


#: RCT experiment that explores how violations of SUTVA impact inference.
RCT = GenericExperiment(
    name="civil_violence_rct",
    description="RCT experiment showing the effect of SUTVA violations on inference.",
    run_method=run_civil_violence)



#!/usr/bin/env python

from __future__ import print_function
try: # try python2
    from urllib2 import urlopen
except: # python3
    from urllib.request import urlopen
import sys
import os
import tempfile
import gzip
import tarfile
import glob
import subprocess
import shutil

try:
    from StringIO import StringIO
except ImportError:
    from io import BytesIO as StringIO

try:
    from subprocess import DEVNULL # py3k
except ImportError:
    import os
    DEVNULL = open(os.devnull, 'wb')

# Set up functions B and S to convert a string to bytes or vice versa. Useful
# for both python 2 and 3 support
if sys.version_info[0] == 2:
    B = lambda s: s
    S = B
else:
    B = lambda s: bytes(s, 'UTF-8')
    S = lambda b: b.decode(encoding='UTF-8')

def getVersionFromFilename(fileName):
    prefix = "simpact-cyan-"
    suffix = ".tar.gz"

    if fileName.startswith(prefix) and fileName.endswith(suffix):
        return fileName[len(prefix):-len(suffix)]

    raise Exception("Can't get version from " + fileName)

def getPackageData(fullFileName):
    if not os.path.exists(fullFileName): # assume we still need to download the file
        print("Downloading " + fullFileName)
        response = urlopen(fullFileName)
        data = response.read()
    else:
        print("Reading file " + fullFileName)
        with open(fullFileName, "rb") as f:
            data = f.read()

    return data

def createBuildAndSourceDir(path):

    workDir = tempfile.mkdtemp(prefix = "simpactbuildtmp")
    print("Work directory is " + workDir)

    if os.path.isdir(path):
        path = os.path.abspath(path)
        cmakeFile = os.path.join(path, "CMakeLists.txt")
        if not os.path.exists(cmakeFile):
            raise Exception("Directory specified, but no CMakeLists.txt file exists there")

        version = "unknownversion"
        for l in open(cmakeFile).readlines():
            vStr = "set(VERSION"
            if vStr in l:
                x = l[l.find(vStr)+len(vStr):]
                version = x.split(")")[0].strip()
                break

        srcDir = path

        os.chdir(workDir)
        os.mkdir("build")
        os.chdir("build")

        buildDir = os.getcwd()

    else:
        version = getVersionFromFilename(os.path.basename(path))

        data = getPackageData(path)
        gf = gzip.GzipFile(fileobj = StringIO(data), mode='r')
        uncompressedData = gf.read()

        tf = tarfile.open(fileobj = StringIO(uncompressedData))

        os.chdir(workDir)
        tf.extractall()

        # Go to the extracted directory
        os.chdir(glob.glob("*")[0])

        srcDir = os.getcwd()

        os.mkdir("build")
        os.chdir("build")

        buildDir = os.getcwd()

    print("Source directory is: ", srcDir)
    print("Building version: ", version)
    print("Build directory is: ", buildDir)

    return (buildDir, srcDir, version, workDir)
   
def compileCMakeIfNecessary(workDir):
    out = S(subprocess.check_output(["cmake", "--version"]))
    if out and len(out.splitlines()) > 0:
        version = [ x for x in map(int, out.splitlines()[0].split()[-1].split(".")) ]
        while len(version) < 4:
            version += [ 0 ]

        print("Current cmake version is {}".format(version))
        if version >= [ 2, 8, 12, 0 ]: # no rebuild necessary
            return "cmake"

    print("Compiling newer cmake")
    
    cmakeUrl = "https://cmake.org/files/v2.8/cmake-2.8.12.2.tar.gz"
    response = urlopen(cmakeUrl)
    data = response.read()

    gf = gzip.GzipFile(fileobj = StringIO(data), mode='r')
    uncompressedData = gf.read()

    tf = tarfile.open(fileobj = StringIO(uncompressedData))

    os.chdir(workDir)
    
    subDir = "cmakebuild"
    os.mkdir(subDir)
    os.chdir(subDir)
    tf.extractall()

    os.mkdir("inst")
    installPrefix = os.path.join(os.getcwd(), "inst")

    # Go to the extracted directory
    os.chdir(glob.glob("cmake*")[0])

    srcDir = os.getcwd()

    os.mkdir("build")
    os.chdir("build")

    subprocess.check_call([ "cmake", "-DCMAKE_BUILD_TYPE=release", "-DCMAKE_INSTALL_PREFIX={}".format(installPrefix), ".."])
    subprocess.check_call([ "make", "install" ])

    return os.path.join(installPrefix, "bin", "cmake")


def gccSupportsOverride(workDir):
    src = """
struct A
{
    virtual void f() { x = 10; }
    int x;
};

struct B : public A
{
    void f() override { x = 11; }
};

int main(void)
{
    B b;
    return 0;
}
"""
    os.chdir(workDir)

    fName = "overriddetest.cpp"
    open(fName, "wt").write(src)
    try:
        subprocess.check_call(["g++", "-c", "-o", fName + ".o", "-std=c++0x", fName])
        return True
    except subprocess.CalledProcessError:
        return False

def gccSupportsSteadyClock(workDir):
    src = """
#include <chrono>

int main(void)
{
    auto t0 = std::chrono::steady_clock::now();
    return 0;
}
"""
    os.chdir(workDir)

    fName = "steadyclocktest.cpp"
    open(fName, "wt").write(src)
    try:
        subprocess.check_call(["g++", "-c", "-o", fName + ".o", "-std=c++0x", fName])
        return True
    except subprocess.CalledProcessError:
        return False

def build(path, extraCMakeOpts):

    buildDir, srcDir, version, workDir = createBuildAndSourceDir(path)

    cmakeExe = compileCMakeIfNecessary(workDir)

    defines = ''
    defines += ' -Doverride=' if not gccSupportsOverride(workDir) else ''
    defines += ' -Dsteady_clock=system_clock' if not gccSupportsSteadyClock(workDir) else ''
    extraOpts = [ '-DCMAKE_CXX_FLAGS={}'.format(defines) ] if defines else [ ]

    print("Using extra CMake options: {}".format(extraOpts))

    os.chdir(buildDir)
    subprocess.check_call([cmakeExe, srcDir] + extraCMakeOpts + extraOpts)
    subprocess.check_call(["make"])

    return (workDir, buildDir, version)

def createProfileDEntry():
    os.mkdir("etc")
    os.mkdir("etc/profile.d")
    open("etc/profile.d/simpactcyan.sh", "wt").write("""

if [ -z "$PYTHONPATH" ] ; then
	PYTHONPATH=/usr/share/simpact-cyan/python
else
	PYTHONPATH="$PYTHONPATH":/usr/share/simpact-cyan/python
fi

export PYTHONPATH
""")

def checkCommandLineArguments():
    try:
        fileName = None
        options = set()
        for o in sys.argv[1:]:
            if o.startswith("--"):
                if o in [ "--noupdate", "--copypackage" ]:
                    options.add(o[2:])
                else:
                    raise Exception("Unknown option '{}'".format(o))
            else:
                if fileName is None:
                    fileName = o
                else:
                    raise Exception("Trying to set filename to '{}' but was already set to '{}'".format(o,fileName))
        
        if fileName is None:
            raise Exception("No path was specified")

    except Exception as e:
        print("""Error: {}

Usage: {} [--noupdate] [--copypackage] path

Here 'path' can be a local .tar.gz file, a .tar.gz file available over http
or the base directory of a project (containing the main CMakeLists.txt file).
""".format(e, sys.argv[0]))
        sys.exit(-1)

    return (fileName, [ o for o in options])




"""Configurations files for the HIV simulator."""

TOY_CONFIG = {
    "population.nummen": 200,
    "population.numwomen": 200,
    "population.simtime": 40,
    "hivtransmission.param.c": 0.10,
}

TOY_INTERVENTION_CONFIG = {
    "population.nummen": 200,
    "population.numwomen": 200,
    "population.simtime": 40,
    "hivtransmission.param.c": 0.1649,
}

# Take from:
# https://rstudio-pubs-static.s3.amazonaws.com/93073_0ce8e66a84184adfae0f7a8075909d1c.html
DEFAULT_CONFIG = {
    "population.nummen": 5000,
    "population.numwomen": 5000,
    "population.eyecap.fraction": 0.1,  # 1
    "syncpopstats.interval": 1,

    # 110 is 10 years of "burn in", followed by 100 years of HIV transmission
    "population.simtime": 110,
    "hivseed.time": 10,
    "hivseed.type": "amount",
    "hivseed.amount": 100,     # 100 out of 10000 is 1% of the population

    # HIV is introduced at an age that the individual is likely already sexually
    # active, in a relationship, and still has lots of time to infect others
    "hivseed.age.min": 20,
    "hivseed.age.max": 25,

    # Nobody will start ART during the simulations
    # This will result in timing of HIV diagnosis way beyond the simulation period.
    "diagnosis.baseline": -100,

    # We are inflating the transmission parameters "for effect"
    "hivtransmission.param.a": -0.3,
    "hivtransmission.param.b": -8,
    "hivtransmission.param.c": 0.1649,
}



"""Average treatment effect estimation with propensity score matching."""
from time import perf_counter

import pandas as pd

from causality.estimation.parametric import PropensityScoreMatching
from whynot.framework import InferenceResult


def estimate_treatment_effect(covariates, treatment, outcome):
    """Estimate treatment effects using propensity score matching.

    Wrapper bound causality.estimation.parametric.PropensityScoreMatching

    Parameters
    ----------
        covariates: `np.ndarray`
            Array of shape [num_samples, num_features] of features
        treatment:  `np.ndarray`
            Binary array of shape [num_samples]  indicating treatment status for each
            sample.
        outcome:  `np.ndarray`
            Array of shape [num_samples] containing the observed outcome for each sample.

    Returns
    -------
        result: `whynot.framework.InferenceResult`
            InferenceResult object for this procedure

    """
    start_time = perf_counter()
    data = pd.DataFrame({'treatment': treatment, 'outcome': outcome})

    covdict = {}
    for j in range(covariates.shape[1]):
        name = str(j)
        data[name] = covariates[:, j]
        covdict[name] = 'c'

    matcher = PropensityScoreMatching()
    ate = matcher.estimate_ATE(data, 'treatment', 'outcome', covdict, bootstrap=True)
    stop_time = perf_counter()

    return InferenceResult(ate=ate[1], stderr=None,
                           ci=(ate[0], ate[2]),
                           individual_effects=None,
                           elapsed_time=stop_time - start_time)



from setuptools import setup, find_packages

setup(name="whynot",
      version="0.10.12",
      author="John Miller",
      author_email="miller_john@berkeley.edu",
      description="A framework for benchmarking causal inference.",
      packages=find_packages(),
      include_package_data=True,
      install_requires=[
          "causality",
          "dataclasses",
          "mesa",
          "numpy",
          "pandas",
          "pyomo",
          "py_mini_racer",
          "scipy",
          "sklearn",
          "statsmodels",
          "tqdm",
          ],
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: MacOS",
          "Operating System :: POSIX",
          ],
      python_requires='>=3.6')



"""Tests for causal inference methods."""
import numpy as np

import whynot as wn


def test_suite():
    """Test whether or not the entire suite runs as expected."""
    num_samples = 200
    num_features = 15

    covariates = np.random.randn(num_samples, num_features)
    treatment = (np.random.rand(num_samples) < 0.5).astype(np.int64)
    outcomes = np.random.randn(num_samples)

    _ = wn.causal_suite(covariates, treatment, outcomes)



#!/usr/bin/env python

from __future__ import print_function
from common import *

def listFilesUnderDirectory(dirName):
    matches = []
    for root, dirnames, filenames in os.walk(dirName):
        for filename in filenames:
            matches.append(os.path.join(root, filename))

    return matches

def getFedoraVersion():
    return "fc" + open("/etc/fedora-release").read().splitlines()[0].split()[2]

def createFedoraPackage(workDir, buildDir, version):
    p = os.path.join(buildDir, "simpact-cyan-release")
    output = S(subprocess.check_output(["ldd", p ]))

    instDir = os.path.join(workDir, "BUILDROOT")
    os.mkdir(instDir)
    os.chdir(buildDir)
    subprocess.check_call([ "make", "install", "DESTDIR={}".format(instDir)])

    os.chdir(instDir)
    createProfileDEntry()

    fcVersion = getFedoraVersion()

    specFile = os.path.join(workDir, "simpactcyan.spec")
    with open(specFile, "wt") as f:
        f.write("""
Name:		simpactcyan
Version:	{}
Release:	1%{{?dist}}
Summary:	Simpact Cyan
Requires:	libtiff gsl
License:	GPL

%description
SimpactCyan programs for individual based simulations using the MNRM algorithm

%files
""".format(version))

        for n in listFilesUnderDirectory(instDir):
            if not n.startswith(instDir):
                raise Exception("Filename {} does not start with {}".format(n, instDir))
        
            f.write(n[len(instDir):] + "\n")

    print(open(specFile).read())

    out = S(subprocess.check_output(["rpmbuild", "-bb", "--buildroot=" + instDir, specFile]))
    for l in out.splitlines():
        pref = "Wrote:"
        if l.startswith(pref):
            return l[len(pref):].strip()

    return None

def main():

    startDir = os.getcwd()
    fullFileName, options = checkCommandLineArguments()

    if not "noupdate" in options:
        subprocess.call([ "sudo", "dnf", "update", "-y" ])
        subprocess.call([ "sudo", "dnf", "install", "-y", "rpm-build" ])
        subprocess.call([ "sudo", "dnf", "install", "-y", "cmake" ])
        subprocess.call([ "sudo", "dnf", "install", "-y", "make" ])
        subprocess.call([ "sudo", "dnf", "install", "-y", "gcc" ])
        subprocess.call([ "sudo", "dnf", "install", "-y", "gcc-c++" ])
        subprocess.call([ "sudo", "dnf", "install", "-y", "gsl-devel"])
        subprocess.call([ "sudo", "dnf", "install", "-y", "libtiff-devel"])

    workDir, buildDir, version = build(sys.argv[1], [ "-DCMAKE_INSTALL_PREFIX=/usr" ])
    outputFileName = createFedoraPackage(workDir, buildDir, version)
    print("Output file is: " + outputFileName)

    if "copypackage" in options:
        shutil.copy(outputFileName, os.path.join(startDir, os.path.basename(outputFileName)))

if __name__ == "__main__":
    main()



"""Base experimental class."""
import copy
import dataclasses
import inspect
import itertools
import random
from typing import Any

import numpy as np
from tqdm.auto import tqdm

from whynot import utils


PARAM_COLLECTION = "__parameter_collection__"


@dataclasses.dataclass
class Dataset:
    # pylint:disable-msg=too-few-public-methods
    """Observational dataset and grouth truth unit-level effects.

    Attributes
    ----------
        covariates: `np.ndarray`
            Array of shape [num_samples, num_features] of covariates for each unit.
        treatments:  `np.ndarray`
            Array of shape [num_samples] indicating treatment status for each unit.
        outcomes:  `np.ndarray`
            Array of shape [num_samples] containing the observed outcome for each unit.
        true_effects: `np.ndarray`
            Array of shape [num_samples] containing the unit-level treatment
            effects, :math:`Y_i(1) - Y_i(0)` for each :math:`i`.
        sate: float
            Sample average treatment effect based on ground truth unit-level effects.

    """

    covariates: np.ndarray
    treatments: np.ndarray
    outcomes: np.ndarray
    true_effects: np.ndarray

    @property
    def sate(self):
        """Return the sample average treatment effect."""
        return np.mean(self.true_effects)

    def bootstrap_sample_ate_ci(self, num_bootstrap_samples=2000):
        """Bootstrap a 95% confidence interval for the sample ate."""
        means = []
        for _ in range(num_bootstrap_samples):
            sample = np.random.choice(
                self.true_effects, size=len(self.true_effects), replace=True)
            means.append(np.mean(sample))
        return (np.quantile(means, 0.025), np.quantile(means, 0.975))


@dataclasses.dataclass
class Run:
    r"""Encapsulate a trajectory from a dynamical system simulator.

    Attributes
    ----------
        states:
            Sequence of states :math:`x_{t_1}, x_{t_2}, \dots` produced by the
            simulator.
        times:
            Sequence of sampled times :math:`{t_1},{t_2}, \dots` at which states
            are recorded.

    Examples
    --------
    >>> state_at_year_2015 = run[2015]

    """

    states: list
    times: list

    def __getitem__(self, time):
        """Return the state at the given time."""
        time_index = np.argmin(np.abs(np.array(self.times) - time))
        return self.states[time_index]

    @property
    def initial_state(self):
        """Return initial state of the run."""
        return self.states[0]


@dataclasses.dataclass
class InferenceResult:
    # pylint:disable-msg=too-few-public-methods
    """Object to store results of causal inference method.

    Attributes
    ----------
        ate:
            Estimated average treatment effect
        stderr:
            Reported standard error of the ATE estimate. Only available
            if supported by the method, otherwise None.
        ci:
            Reported 95% confidence interval for the ATE (lower_bound, upper_bound).
            Only available if supported by the method, otherwise None.
        individual_effects:
            Heterogeneous treatment effect for each unit.
            Only available if supported by the method, otherwise None.
        elapsed_time:
            How long in second (wall-clock time) it took to produce the estimate.

    """

    ate: float = None
    stderr: float = None
    ci: tuple = (None, None)  # pylint: disable-msg=invalid-name
    individual_effects: np.ndarray = None
    elapsed_time: float = None


@dataclasses.dataclass
class ExperimentParameter:
    # pylint:disable-msg=too-few-public-methods
    """Container for a parameter to vary in an experiment.

    Attributes
    ----------
        name:
            Name of the parameter
        description:
            Parameter description
        default:
            Default (uninitialized) value of the parameter
        values:
            Iterator of parameter values that supports sampling (for random search).

    """

    name: str
    default: Any
    values: Any = (None,)
    description: str = ""


def parameter(name, default, values=(None,), description=""):
    """Decorate functions in an experiment with parameters.

    Usage:
        @parameter(name="treatment-bias", default=0.2, values=[0.1, 0.2, 0.5, 0.9],
                   description="How much is treatment biased for control group.")
        def propensity_score(control_run, treatment_bias)
            ....
    """
    exp_param = ExperimentParameter(name, default, values, description)

    def parameter_decorator(func):
        """Attach parameter collection object to the function."""
        # Ensure the parameter is an argument to the function
        method_params = inspect.signature(func).parameters
        if name not in method_params:
            raise ValueError(f"{name} is not an argument to {func.__name__}")

        if hasattr(func, PARAM_COLLECTION):
            getattr(func, PARAM_COLLECTION).add_parameter(exp_param)
        else:
            setattr(func, PARAM_COLLECTION, ParameterCollection([exp_param]))

        return func

    return parameter_decorator


def extract_params(func, standard_args):
    """Return WhyNot parameters for user-defined function.

    Performs error-checking to ensure parameters are disjoint from
    standard arguments and all arguments to the function are either
    standard arguments or parameters.

    Parameters
    ----------
        func: function
            Possibly parameterized function.

        standard_args: list
            A list of possible arguments provided by the calling class itself
            and shouldn't be treated as parameters for the function.

    Returns
    -------
        params: `whynot.framework.ParameterCollection`
            A collection of parameters for the func.

    """
    if not callable(func):
        msg = f"Trying to extract parameters from {func.__name__}, but not callable."
        raise ValueError(msg)

    # Extract parameters specified by the user via the @parameter decorator.
    specified_params = ParameterCollection([])
    if hasattr(func, PARAM_COLLECTION):
        specified_params = getattr(func, PARAM_COLLECTION)

    # Ensure standard_args is disjoint from the specified params.
    for arg in standard_args:
        if arg in specified_params:
            msg = (f"{arg} is both a parameter and a standard argument to {func.__name__}.")
            raise ValueError(msg)

    # By construction, every element in specified_params
    # must appear in the function signature, i.e. method_params.
    method_args = inspect.signature(func).parameters
    for arg in method_args:
        if arg not in standard_args and arg not in specified_params:
            raise ValueError(f"{arg} in signature, but not specified.")

    return specified_params


class ParameterCollection():
    """Lightweight wrapper class around a set of parameters.

    Provides utility functions to support sampling and assigning subsets of the
    parameters.

    Enforces name uniqueness. Every parameter should have a unique name.

    """

    def __init__(self, params=None):
        """Params is a list of Parameter objects."""
        self.params = {}
        if params is not None:
            for param in params:
                if param.name in self.params:
                    raise ValueError(f"Duplicate name {param.name}")
                self.params[param.name] = param

    def add_parameter(self, param):
        """Add a new parameter to the existing collection."""
        if param.name in self.params:
            raise ValueError(f"Adding parameter with duplicate name {param.name}")
        self.params[param.name] = param

    def default(self):
        """Return the default parameter setting for each parameter."""
        return dict((name, p.default) for name, p in self.params.items())

    def sample(self):
        """Return a random parameter setting for each parameter."""
        return dict((name, random.choice(p.values)) for name, p in self.params.items())

    def project(self, specified_params):
        """Return fully instantiated parameters using defaults for unspecified params.

        specified_params is a dict of the arguments passed to the run method.
        Any unspecified params are set to their default values.

        Returns a dictonary params[param_name] -> param_value
        """
        for param_name in specified_params:
            if param_name not in self.params:
                raise ValueError(f"Parameter {param_name} specified, but not used!.")

        params = {}
        for name, param in self.params.items():
            if name in specified_params:
                params[name] = specified_params[name]
            else:
                params[name] = param.default

        return params

    def __contains__(self, name):
        """Check if param name is specified."""
        return name in self.params

    def __getitem__(self, name):
        """Return the parameter object corresponding to name."""
        if name in self.params:
            return self.params[name]
        raise ValueError(f"{name} not found")

    def __add__(self, collection):
        """Add two parameter collections together."""
        original_params = list(self.params.values())
        new_params = list(collection.params.values())
        return ParameterCollection(original_params + new_params)

    def __iter__(self):
        """Iterate over the collection."""
        for param in self.params.values():
            yield param

    def __repr__(self):
        """See ___str___."""
        return self.__str__()

    def __str__(self):
        """Display the collection in a human readable format.

        For instance:
        Params:
            Name:		    hidden_dim
            Description:	hidden dimension of 2-layer ReLu network response.
            Default:	    32
            Values:		    [8, 16, 32, 64, 128, 256, 512]
        """
        class_display = "Params:"
        param_display = "\tName:\t\t{}\n\tDescription:\t{}\n\tDefault:\t{}\n\tValues:\t\t{}\n"
        for param_ in self.params.values():
            class_display += "\n" + param_display.format(
                param_.name, param_.description, param_.default,
                list(param_.values))
        return class_display

    def grid(self):
        """Return a parameter grid."""
        grid = []
        for values in itertools.product(*[p.values for p in self.params.values()]):
            settings = {}
            for param_, value in zip(self.params, values):
                settings[param_.name] = value
            grid.append(settings)
        return grid


####################
# Generic Simulator
####################
class GenericExperiment():
    # pylint:disable-msg=too-few-public-methods
    """Encapsulate a causal simulation experiment."""

    def __init__(self, name, description, run_method):
        """Initialize a generic experiment class."""
        self.name = name
        self.description = description
        self.run_method = run_method

        run_args = ["num_samples", "seed", "parallelize", "show_progress"]
        self.params = extract_params(run_method, run_args)

    def get_parameters(self):
        """Return parameters of the experiment."""
        return self.params

    def run(self, num_samples, seed, parallelize=True, show_progress=False, **parameter_args):
        """Run the experiment and return a causal dataset."""
        run_parameters = self.params.project(parameter_args)
        results = self.run_method(num_samples=num_samples, seed=seed,
                                  show_progress=show_progress,
                                  parallelize=parallelize, **run_parameters)

        (covariates, treatment, outcome), ground_truth = results
        return Dataset(covariates=covariates, treatments=treatment,
                       outcomes=outcome, true_effects=ground_truth)


#############################
# Dynamical system simulators
#############################
class DynamicsExperiment():
    """Encapsulate a causal experiment on a system dynamics simulator."""

    # pylint:disable-msg=too-many-arguments
    def __init__(self, name, description, simulator_config, intervention,
                 state_sampler, simulator, propensity_scorer,
                 outcome_extractor, covariate_builder):
        """Create an experiment based on system dynamics simulator.

        Parameters
        ----------
            name: str
                Name of the experiment
            description: str
                Short description of the experiment
            simulator:
                WhyNot systems dynamics simulator
            simulator_config: `Config`
                Instantiated simulator `Config` for all runs
            intervention: `Intervention`
                Instantiated simulator `Intervention` for treatment runs
            state_sampler:
                Function to sample initial simulator `State` (s)
            propensity_score:
                Either float or function to generate propensity scores for a set of rollouts
            outcome_extractor:
                Function to extract outcome measurement from rollouts
            covariate_builder:
                Function to build covariates from rollouts.

        All of ``simulator_config``, ``intervention``, ``state_sampler``,
        ``propensity_scorer``, ``outcome_extractor``, and ``covariate_builder``
        support a rich set of patterns. We describe the expected API below.

        Examples
        --------
        First, ``simulator_config`` can either be a constant ``Config`` object, or a,
        potentially parameterized, function that returns a ``Config``. For
        instance:

        .. code-block::

            # simulator_config can be either a constant or a function that returns a
            DynamicsExperiment(
                config=Config(123),...)

            @parameter(name="p", default=0.2)
            def config(p):
                return Config(1234, parameter=p)

        Similarly, ``intervention`` can either be a constant ``Intervention``
        object, or a, potentially parameterized, function that returns an
        ``Intervention``.

        .. code-block::

                DynamicsExperiment(
                    intervention=Intervention(year=92),...)

                @parameter(name="p", default=0.2)
                def intervention(p):
                    return Intervention(year=92, parameter=p)

        The ``state_sampler`` generates a sequence of initial states. If the
        sampler uses randomness, it must include ``rng`` in the signature and use
        it as the sole source of randomness. If ``num_samples`` is in the
        signature, then the ``state_sampler`` must returns a sequence of
        ``num_samples`` states.  Otherwise, it must returns a single state.

        .. code-block::

            def state_sample(rng):
                # Return a single random sample, use randomness in rng
                return State(rng.rand())

            def state_sampler(rng, num_samples):
                # Return a list of initial states
                return [State(rng.rand()) for _ in range(num_samples)]

        The ``propensity_scorer`` produces, for each run, the probability of
        being assigned to treatment in the observational dataset. The
        ``propensity_scorer`` can be a constant, in which case the probability
        of treatment is uniform for all units. Otherwise, if
        ``propensity_scorer`` is a function, it either produces a propensity
        score for each run separately or all together (for correlated treatment
        assignment).

        For individual runs, the function may include one or more
        of ``run``, ``control_run``, or ``treatment_run`` in the signature.  For
        a whole population, use ``control_runs``, ``runs``, or
        ``treatment_runs`` in the function signature. The ``run`` argument
        defaults to ``control_run`` and similarly for ``runs``. Both ``config`` and
        ``intervention`` can be passed as argument to access the
        ``simulator_config`` and ``intervention`` for the given experiment.

        .. code-block::

            # self.propensity_scorer can be a constant
            DynamicsExperiment(
                propensity_scorer=0.8),
                ...)

            ##
            # Compute treatment assignment probability for each run separately.
            ##
            def propensity_scorer(control_run):
                # Return propensity score for a single run, based on control run
                return 0.5 if control_run[20].values()[0] > 0.2 else 0.1

            def propensity_scorer(run):
                # Argument run defaults to the control_run
                return 0.5 if run[20].values()[0] > 0.2 else 0.1

            def propensity_scorer(run, config, intervention):
                # You can pass the config and intervention
                return 0.5 if run[intervention.year].values()[0] > 0.2 else 0.1

            def propensity_scorer(treatment_run):
                # Return propensity score for a single run, based on treatment run
                return 0.5 if treatment_run[20].values()[0] > 0.2 else 0.1

            def propensity_scorer(treatment_run, control_run, config):
                # Assign treatment based on runs with both settings.
                return 0.2

            ##
            # Compute treatment assignment probability for all runs
            # simultaneously. This allows for correlated propensity scores.
            ##
            def propensity_scorer(control_runs):
                # Return propensity score for all runs, based on control run
                return [0.1  for _ in control_runs]

            def propensity_scorer(runs):
                # runs defaults to control_runs
                return [0.1  for _ in runs]

            etc..

        The ``outcome_extractor`` returns, for a given run, the outcome
        :math:`Y`. Both ``config`` and ``intervention`` can be passed as
        arguments.  For instance,

        .. code-block::

            def outcome_extractor(run):
                # Return the first state coordinate at time step 100
                return run[100].values()[0]

            def outcome_extractor(run, config, intervention):
                # You can also specify one (or both) of config/intervention
                # First state coordinate 20 steps after intervention.
                return run[intervention.time + 20].values()[0]

        Finally, the ``covariate_builder`` returns, for a given run, the
        observed covariates :math:`X`. If the signature includes ``runs``, the
        method returns covariates for the entire sequence of runs. Otherwise, it
        produces covariates for a single time step. Both ``config`` and
        ``intervention`` can be passed.

        .. code-block::

            def covariate_builder(run):
                # Functions with the argument `run` return covariates for a
                # single rollout.
                return run.initial_state.values()

            def covariate_builder(runs):
                # Functions with the argument `runs` return covariates for the
                # entire sequence of runs.
                return np.array([run.initial_state.value() for run in runs])

            def covariate_builder(config, intervention, run):
                # Can optionally specify `config` or `intervention` in both cases.
                return run[intervention.year].values()

        """
        self.name = name
        self.description = description
        self.simulator = simulator

        if not callable(simulator_config):
            self.simulator_config = lambda: simulator_config
        else:
            self.simulator_config = simulator_config

        if not callable(intervention):
            self.intervention = lambda: intervention
        else:
            self.intervention = intervention

        self.state_sampler = state_sampler

        if not callable(propensity_scorer):
            if isinstance(propensity_scorer, float) and 0. <= propensity_scorer <= 1.:
                self.propensity_scorer = lambda: propensity_scorer
            else:
                msg = "If propensity scorer is not callable, it must be a float in [0., 1.]"
                raise ValueError(msg)
        else:
            self.propensity_scorer = propensity_scorer

        self.outcome_extractor = outcome_extractor
        self.covariate_builder = covariate_builder
        self.parameter_collection = self.get_parameters()

    def get_parameters(self):
        """Inspect provided methods and gather parameters.

        Returns
        -------
            params: `whynot.framework.ParameterCollection`
                Collection of all of the parameters specified in the experiment.

        """
        params = ParameterCollection()

        config_args = []
        params += extract_params(self.simulator_config, config_args)
        params += extract_params(self.intervention, config_args)

        initial_state_args = ["rng", "num_samples"]
        params += extract_params(self.state_sampler, initial_state_args)

        propensity_args = ["config", "intervention", "control_run",
                           "control_runs", "treatment_run", "treatment_runs",
                           "run", "runs"]
        params += extract_params(self.propensity_scorer, propensity_args)

        outcome_args = ["config", "intervention", "run"]
        params += extract_params(self.outcome_extractor, outcome_args)

        covariate_args = ["config", "intervention", "run",
                          "runs", "control_run", "control_runs",
                          "treatment_run", "treatment_runs"]
        params += extract_params(self.covariate_builder, covariate_args)

        return params

    @staticmethod
    def _get_args(func):
        """Return the arguments to the function."""
        return inspect.signature(func).parameters

    @staticmethod
    def run_dynamics_simulator(simulate, config, intervention, initial_state, seed):
        """Run the simulation with a common initial state for both treatment and control.

        Use the same initial seed for each run, and return the outcomes of both runs.
        This function is a staticmethod since self is not pickable when the
        class has lambda functions as attributes, e.g. the propensity scorer.

        Parameters
        ----------
            simulate: function
                Entry point to the dynamical systems simulator.
            config:   `Config`
                Configuration object for the simulator, used for both treatment and control.
            intervention: `Intervention`
                Intervention object for the simulator, used for the treatment group.
            initial_state: `State`
                State object used to initial both treatment and control runs of the simulator.
            seed: int
                Random seed used to initialize both runs of the simulator.

        Returns
        -------
            (control_run, treatment_run): `whynot.framework.Run`
                Tuple containing runs of the simulator with and without treatment.

        """
        # Work with copies of the initial state to avoid issues where the
        # simulator had side effects on the state.
        control_run = simulate(copy.deepcopy(initial_state), config=config,
                               intervention=None, seed=seed)
        treatment_run = simulate(copy.deepcopy(initial_state), config=config,
                                 intervention=intervention, seed=seed)
        return control_run, treatment_run

    def _run_all_simulations(self, initial_states, config,
                             intervention, rng, parallelize, show_progress):
        """Run simulation for treatment and control groups for each initial state.

        Parameters
        ----------
            initial_state: list
                List of initial state objects
            config: `Config`
                Configuration object for all runs of the simulator.
            intervention: `Intervention`
                Intervention object for treatment runs of the simulator.
            rng: `np.random.RandomState`
                Source of randomness. Used to choose seeds for different simulator runs.
            parallelize: bool
                Whether or not to execute runs of the simulator in parallel or sequentially.
            show_progress: bool
                Whether or not to display a progress bar for simulator execution.

        Returns
        -------
            control_runs: `list[whynot.framework.Run]`
                List of simulator runs in control group, one for each initial state.
            treatment_runs: `list[whynot.framework.Run]`
                List of simulator runs in treatment group, one for each initial state.

        """
        parallel_args = []
        for state in initial_states:
            seed = rng.randint(0, 9999)
            parallel_args.append((self.simulator.simulate, config, intervention, state, seed))

        if parallelize:
            runs = utils.parallelize(self.run_dynamics_simulator, parallel_args,
                                     show_progress=show_progress)
        elif show_progress:
            runs = [self.run_dynamics_simulator(*args) for args in tqdm(parallel_args)]
        else:
            runs = [self.run_dynamics_simulator(*args) for args in parallel_args]

        control_runs, treatment_runs = list(zip(*runs))

        return control_runs, treatment_runs

    def construct_covariates(self, config, intervention, control_runs,
                             treatment_runs, treatments, params):
        """Build the observational dataset after running simulator.

        Parameters
        ----------
            config: `Config`
                Config object used when running the simulator.
            intervention: `Intervention`
                Intervention object used when running the simulator with treatment.
            control_runs: `list[whynot.framework.Run]`
                List of all of the simulator runs in control condition.
            treatment_runs: `list[whynot.framework.Run]`
                List of all of the simulator runs with treatment applied.
            treatments: `list[bool]`
                Whether or not a particular run was assigned to treatment or control.
            params: dict
                Dictionary mapping parameter names to assigned values.

        Returns
        -------
            covariates: `np.ndarray`
                Numpy array of shape [num_samples, num_features] containing the
                covariates for the observational dataset.

        """
        observed_runs = []
        for treatment, control_run, treatment_run in zip(treatments, control_runs, treatment_runs):
            run = treatment_run if treatment else control_run
            observed_runs.append(run)

        covariate_builder_args = self._get_args(self.covariate_builder)
        kwargs = {}
        for arg in covariate_builder_args:
            if arg in params:
                kwargs[arg] = params[arg]

        if "config" in covariate_builder_args:
            kwargs["config"] = config

        if "intervention" in covariate_builder_args:
            kwargs["intervention"] = intervention

        # Generate all of the covariates in batch
        if "runs" in covariate_builder_args:
            dataset = self.covariate_builder(runs=observed_runs, **kwargs)
        else:
            # Generate the covariates one by one

            dataset = np.array([self.covariate_builder(run=observed_run, **kwargs)
                                for observed_run in observed_runs])

        if dataset.ndim == 1:
            dataset = np.expand_dims(dataset, axis=1)
        return dataset

    def get_outcomes(self, config, intervention, control_runs, treatment_runs, params):
        """Return outcomes for both treatment and control runs.

        Parameters
        ----------
            config: `Config`
                Simulator configuration object used to generate the runs.
            intervention: `Intervention`
                Simulator intervention used to specify treatment runs.
            control_runs:   `list[whynot.framework.Run]`
                List of simulator runs without intervention.
            treatments_runs: `list[whynot.framework.Run]`
                List of simulator runs with intervention.
            params:
                Dictionary mapping parameter names to arguments.

        Returns
        -------
            control_outcomes: `np.ndarray`
                Array of shape [num_samples] with control outcome :math:`Y_i(0)` for each i.
            treatment_outcomes: `np.ndarray`
                Array of shape [num_samples] with treatment outcome :math:`Y_i(1)` for each i.

        """
        extractor_args = self._get_args(self.outcome_extractor)
        kwargs = {}
        for arg in extractor_args:
            if arg in params:
                kwargs[arg] = params[arg]
        if "config" in extractor_args:
            kwargs["config"] = config
        if "intervention" in extractor_args:
            kwargs["intervention"] = intervention

        def get_outcome(run):
            return self.outcome_extractor(run=run, **kwargs)

        control_outcomes = np.array([get_outcome(run) for run in control_runs])
        treatment_outcomes = np.array([get_outcome(run) for run in treatment_runs])

        return control_outcomes, treatment_outcomes

    def get_propensity_scores(self, config, intervention, control_runs, treatment_runs, params):
        """Return probability of treatment for a set of runs.

        Parameters
        ----------
            config: `Config`
                Simulator configuration object used to generate the runs.
            intervention: `Intervention`
                Simulator intervention used to specify treatment runs.
            control_runs:   `list[whynot.framework.Run]`
                List of simulator runs without intervention.
            treatments_runs: `list[whynot.framework.Run]`
                List of simulator runs with intervention.
            params:
                Dictionary mapping parameter names to arguments.

        Returns
        -------
            propensity_score: `np.ndarray`
                Array of shape [num_samples] with probability of treatment for each unit i.

        """
        # pylint:disable-msg=too-many-branches
        if not callable(self.propensity_scorer):
            if isinstance(self.propensity_scorer, float):
                uniform_propensity = self.propensity_scorer
                if 0. <= uniform_propensity <= 1.:
                    return uniform_propensity * np.ones(len(control_runs))

            raise ValueError("If propensity scorer is not callable, it must be a float in [0., 1.]")

        scorer_args = self._get_args(self.propensity_scorer)

        # Build kwargs for scorer based on passed parameters
        kwargs = {}
        for arg in scorer_args:
            if arg in params:
                kwargs[arg] = params[arg]

        if "config" in scorer_args:
            kwargs["config"] = config

        if "intervention" in scorer_args:
            kwargs["intervention"] = intervention

        # Check whether or not the propensity scorer generates all scores at once
        one_shot = False
        if "treatment_runs" in scorer_args:
            one_shot = True
            kwargs["treatment_runs"] = treatment_runs
        if "control_runs" in scorer_args:
            one_shot = True
            kwargs["control_runs"] = control_runs
        if "runs" in scorer_args:
            one_shot = True
            kwargs["runs"] = control_runs

        if one_shot:
            return self.propensity_scorer(**kwargs)

        # If not, then we need to score each run separately.
        propensity_scores = []
        for control_run, treatment_run in zip(control_runs, treatment_runs):
            if "control_run" in scorer_args and "treatment_run" in scorer_args:
                score = self.propensity_scorer(control_run=control_run,
                                               treatment_run=treatment_run,
                                               **kwargs)
            elif "control_run" in scorer_args:
                score = self.propensity_scorer(control_run=control_run, **kwargs)
            elif "treatment_run" in scorer_args:
                score = self.propensity_scorer(treatment_run=treatment_run, **kwargs)
            elif "run" in scorer_args:
                score = self.propensity_scorer(run=control_run, **kwargs)
            else:
                score = self.propensity_scorer(**kwargs)
            propensity_scores.append(score)

        return np.array(propensity_scores)

    def sample_initial_states(self, rng, num_samples, params):
        """Sample initial states for simulator.

        Parameters
        ----------
            rng: `np.random.RandomState`
                Random number generator used for all experiment randomness
            num_samples: int
                Number of samples to use for the experiment.
            params:
                Dictionary mapping parameter names to assigned values.

        Returns
        -------
            initial_states: `list[States]`
                List of length `num_samples` of simulator initial `States.

        """
        state_sampler_args = self._get_args(self.state_sampler)

        # Build kwargs
        kwargs = {}
        for arg in state_sampler_args:
            if arg in params:
                kwargs[arg] = params[arg]

        if "rng" in state_sampler_args:
            kwargs["rng"] = rng

        if "num_samples" in state_sampler_args:
            return self.state_sampler(num_samples=num_samples, **kwargs)

        return [self.state_sampler(**kwargs) for _ in range(num_samples)]

    def get_config(self, params):
        """Return the simulator config.

        Parameters
        ----------
            params: dict
                Dictionary mapping experiment parameter names to assigned values

        Returns
        -------
            config: `Config`
                Simulator configuration object used for this experiment.

        """
        if not callable(self.simulator_config):
            return self.simulator_config

        control_args = self._get_args(self.simulator_config)
        kwargs = {}
        for arg in control_args:
            if arg in params:
                kwargs[arg] = params[arg]

        return self.simulator_config(**kwargs)

    def get_intervention(self, params):
        """Return the intervention correspond to treatment.

        Parameters
        ----------
            params: dict
                Dictionary mapping experiment parameter names to assigned
                values.

        Returns
        -------
            intervention: `Intervention`
                Simulator Intervention object used for treatment group.

        """
        if not callable(self.intervention):
            return self.intervention

        treatment_args = self._get_args(self.intervention)
        kwargs = {}
        for arg in treatment_args:
            if arg in params:
                kwargs[arg] = params[arg]

        return self.intervention(**kwargs)

    def run(self, num_samples, seed=None, parallelize=True, show_progress=False,
            **parameter_args):
        # pylint:disable-msg=too-many-locals
        """Run a basic parameterized experiment on a dynamical system simulator.

        Parameters
        ----------
            num_samples: int
                Number of units to sample and simulate
            show_progress: bool
                Should progress bar be shown?
            parallelize: bool
                If true, the experiment class will run all of the simulations in
                parallel.
            seed: int
                Random number generator seed used to set all internal randomness.
            **parameter_args:
                If the experiment is parameterized, additional arguments to
                select the parameters of a particular run.

        Returns
        -------
            dataset: whynot.framework.Dataset
                Dataset object encapsulating the covariates, treatment assignments, and
                outcomes observed in the experiment, as well as unit-level ground truth.
                The covariates are an array of size [num_samples, num_features] where
                num_features is determined by the covariate builder.

        """
        run_parameters = self.parameter_collection.project(parameter_args)

        rng = np.random.RandomState(seed)

        config = self.get_config(run_parameters)
        intervention = self.get_intervention(run_parameters)

        initial_states = self.sample_initial_states(rng, num_samples, run_parameters)

        # Run the model
        control_runs, treatment_runs = self._run_all_simulations(
            initial_states, config, intervention, rng, parallelize, show_progress)

        # Assign treatment
        propensities = self.get_propensity_scores(
            config, intervention, control_runs, treatment_runs, run_parameters)
        treatment = (rng.uniform(size=propensities.shape) < propensities).astype(np.int64)

        # Build observational dataset
        covariates = self.construct_covariates(
            config, intervention, control_runs, treatment_runs, treatment, run_parameters)

        # Extract outcomes
        control_outcomes, treatment_outcomes = self.get_outcomes(
            config, intervention, control_runs, treatment_runs, run_parameters)

        outcomes = np.copy(control_outcomes)
        outcomes[treatment == 1.0] = treatment_outcomes[treatment == 1.0]

        # Ground truth effect is Y(1) - Y(0)
        treatment_effects = treatment_outcomes - control_outcomes

        return Dataset(covariates=covariates, treatments=treatment,
                       outcomes=outcomes, true_effects=treatment_effects)



#!/usr/bin/env python

from __future__ import print_function
from common import *

def getArchitecture():
    out = S(subprocess.check_output(["uname", "-a"]))
    if "x86_64" in out:
        return "amd64"
    return "i386"

def getLibraryPackage(libName):
    out = S(subprocess.check_output(["dpkg-query", "-S", libName]))
    for l in out.splitlines():
        parts = l.split()
        path = parts[1]
        print(parts)
        if os.path.exists(path) and not os.path.islink(path):
            pack = l.split(":")[0]
            print("Package is: " + pack)
            return pack

    raise Exception("Package for {} not found".format(libName))

def getPlatformAndVersion():
    out = S(subprocess.check_output(["lsb_release", "-a"], stderr=DEVNULL))
    platform = "unknown"
    platformVersion = "unknown"
    for l in out.splitlines():
        if "Distributor ID" in l:
            platform = l.split("\t")[1].lower()
        elif "Release" in l:
            platformVersion = l.split("\t")[1].lower()

    return (platform, platformVersion)

def createDebianPackage(workDir, buildDir, version):
    p = os.path.join(buildDir, "simpact-cyan-release")
    output = S(subprocess.check_output(["ldd", p ]))

    libTiff = getLibraryPackage("libtiff.so")
    libGsl = getLibraryPackage("libgsl.so")
    arch = getArchitecture()

    print("TIFF package: " + libTiff)
    print("GSL package: " + libGsl)
    print("Architecture: " + arch)

    instDir = os.path.join(workDir, "fakeinst")
    os.mkdir(instDir)
    os.chdir(buildDir)
    subprocess.check_call([ "make", "install", "DESTDIR={}".format(instDir)])

    os.chdir(instDir)
    createProfileDEntry()

    origFiles = glob.glob("*")
    lines = S(subprocess.check_output(["du" ] + origFiles + [ "--max-depth=0" ])).splitlines()
    totalSize = 0
    for l in lines:
        totalSize += int(l.split()[0])

    print("Total size: ", totalSize)

    subprocess.check_call(["tar", "cfzv", "data.tar.gz" ] + origFiles)
    
    # Remove the original files themselves
    for i in origFiles:
        subprocess.check_call([ "rm", "-rf", i ])

    open("debian-binary", "wt").write("2.0\n")
    open("control", "wt").write("""Package: simpactcyan
Version: {}
Section: user
Priority: optional
Architecture: {}
Installed-Size: {}
Maintainer: Jori Liesenborgs <jori.liesenborgs@uhasselt.be>
Description: SimpactCyan programs for individual based simulations using the MNRM algorithm
Depends: {}, {}, libgomp1
""".format(version, arch, totalSize, libGsl, libTiff))

    subprocess.check_call(["tar", "cfzv", "control.tar.gz", "control"])
    subprocess.check_call(["ar", "-r", "simpactcyan.deb", "debian-binary", "control.tar.gz", "data.tar.gz" ])

    platform, platformVersion = getPlatformAndVersion()

    outputFileName = "simpactcyan-{}-{}-{}-{}.deb".format(version, platform, platformVersion, arch)
    shutil.move("simpactcyan.deb", outputFileName)

    return outputFileName

def main():

    startDir = os.getcwd()
    fullFileName, options = checkCommandLineArguments()

    if not "noupdate" in options:
        subprocess.call([ "sudo", "apt-get", "update", "-y" ])
        # subprocess.call([ "sudo", "apt-get", "upgrade", "-y" ])
        subprocess.call([ "sudo", "apt-get", "install", "-y", "cmake", "make", "gcc", "g++"])
        subprocess.call([ "sudo", "apt-get", "install", "-y", "libgsl-dev"])
        subprocess.call([ "sudo", "apt-get", "install", "-y", "libgsl0-dev"])
        subprocess.call([ "sudo", "apt-get", "install", "-y", "libtiff-dev"])
        subprocess.call([ "sudo", "apt-get", "install", "-y", "libtiff4-dev"])
        subprocess.call([ "sudo", "apt-get", "install", "-y", "libtiff5-dev"])

    # We also disable the 'override' compiler check this way, so that it also
    # builds using an older compiler

    workDir, buildDir, version = build(fullFileName, [ "-DCMAKE_INSTALL_PREFIX=/usr" ])
    outputFileName = createDebianPackage(workDir, buildDir, version)
    print("Output file is: " + outputFileName)

    if "copypackage" in options:
        shutil.copy(outputFileName, os.path.join(startDir, os.path.basename(outputFileName)))

if __name__ == "__main__":
    main()
